@@ -22,6 +22,26 @@ from shutil import rmtree
 from sys import argv, exit
 from tempfile import mkdtemp
 
+from pkg_resources import require, VersionConflict, DistributionNotFound
+
+# We don't admit our dependency on pip in setup.py, lest a naive user simply
+# say `pip install peep.tar.gz` and thus pull down an untrusted copy of pip
+# from PyPI. Instead, we make sure it's installed and new enough here and spit
+# out an error message if not:
+def activate(specifier):
+    """Make a compatible version of pip importable. Raise a RuntimeError if we
+    couldn't."""
+    try:
+        for distro in require(specifier):
+            distro.activate()
+    except (VersionConflict, DistributionNotFound):
+        raise RuntimeError('The installed version of pip is too old; peep '
+                           'requires ' + specifier)
+
+activate('pip>=0.6.2')  # Before 0.6.2, the log module wasn't there, so some
+                         # of our monkeypatching fails. It probably wouldn't be
+                         # much work to support even earlier, though.
+
 import pip
 from pip.log import logger
 from pip.req import parse_requirements
@@ -208,11 +228,20 @@ def hashes_of_requirements(requirements):
 def hash_mismatches(expected_hash_map, downloaded_hashes):
     """Yield the list of allowed hashes, package name, and download-hash of
     each package whose download-hash didn't match one allowed for it in the
-    requirements file."""
+    requirements file.
+
+    If a package is missing from ``download_hashes``, ignore it; that means
+    it's already installed and we're not risking anything.
+
+    """
     for package_name, expected_hashes in expected_hash_map.iteritems():
-        hash_of_download = downloaded_hashes[package_name]
-        if hash_of_download not in expected_hashes:
-            yield expected_hashes, package_name, hash_of_download
+        try:
+            hash_of_download = downloaded_hashes[package_name]
+        except KeyError:
+            pass
+        else:
+            if hash_of_download not in expected_hashes:
+                yield expected_hashes, package_name, hash_of_download
 
 
 def peep_hash(argv):
@@ -237,6 +266,17 @@ def peep_hash(argv):
         return COMMAND_LINE_ERROR
 
 
+class EmptyOptions(object):
+    """Fake optparse options for compatibility with pip<1.2
+
+    pip<1.2 had a bug in parse_requirments() in which the ``options`` kwarg
+    was required. We work around that by passing it a mock object.
+
+    """
+    default_vcs = None
+    skip_requirements_regex = None
+
+
 def peep_install(argv):
     """Perform the ``peep install`` subcommand, returning a shell status code
     or raising a PipException.
@@ -246,20 +286,25 @@ def peep_install(argv):
     """
     req_paths = list(requirement_args(argv, want_paths=True))
     if not req_paths:
-        print "You have to specify one or more requirements files with the -r option, because otherwise there's nowhere for peep to look up the hashes."
+        print "You have to specify one or more requirements files with the -r option, because"
+        print "otherwise there's nowhere for peep to look up the hashes."
         return COMMAND_LINE_ERROR
 
     # We're a "peep install" command, and we have some requirement paths.
-
-    requirements = list(chain(*(parse_requirements(path) for
-                                path in req_paths)))
-    downloaded_hashes, downloaded_versions = {}, {}
+    requirements = list(chain(*(parse_requirements(path,
+                                                   options=EmptyOptions())
+                                for path in req_paths)))
+    downloaded_hashes, downloaded_versions, satisfied_reqs = {}, {}, []
     with ephemeral_dir() as temp_path:
         for req in requirements:
-            name = req.req.project_name
-            archive_filename = pip_download(req, argv, temp_path)
-            downloaded_hashes[name] = hash_of_file(join(temp_path, archive_filename))
-            downloaded_versions[name] = version_of_archive(archive_filename, name)
+            req.check_if_exists()
+            if req.satisfied_by:  # This is already installed.
+                satisfied_reqs.append(req)
+            else:
+                name = req.req.project_name
+                archive_filename = pip_download(req, argv, temp_path)
+                downloaded_hashes[name] = hash_of_file(join(temp_path, archive_filename))
+                downloaded_versions[name] = version_of_archive(archive_filename, name)
 
         expected_hashes, missing_hashes = hashes_of_requirements(requirements)
         mismatches = list(hash_mismatches(expected_hashes, downloaded_hashes))
@@ -271,7 +316,9 @@ def peep_install(argv):
 
         # Mismatched hashes:
         if mismatches:
-            print "THE FOLLOWING PACKAGES DIDN'T MATCHES THE HASHES SPECIFIED IN THE REQUIREMENTS FILE. If you have updated the package versions, update the hashes. If not, freak out, because someone has tampered with the packages.\n"
+            print "THE FOLLOWING PACKAGES DIDN'T MATCHES THE HASHES SPECIFIED IN THE REQUIREMENTS"
+            print "FILE. If you have updated the package versions, update the hashes. If not,"
+            print "freak out, because someone has tampered with the packages.\n"
         for expected_hashes, package_name, hash_of_download in mismatches:
             hash_of_download = downloaded_hashes[package_name]
             preamble = '    %s: expected%s' % (
@@ -285,7 +332,9 @@ def peep_install(argv):
 
         # Missing hashes:
         if missing_hashes:
-            print 'The following packages had no hashes specified in the requirements file, which leaves them open to tampering. Vet these packages to your satisfaction, then add these "sha256" lines like so:\n'
+            print 'The following packages had no hashes specified in the requirements file, which'
+            print 'leaves them open to tampering. Vet these packages to your satisfaction, then'
+            print 'add these "sha256" lines like so:\n'
         for package_name in missing_hashes:
             print '# sha256: %s' % downloaded_hashes[package_name]
             print '%s==%s\n' % (package_name,
@@ -297,6 +346,14 @@ def peep_install(argv):
             return SOMETHING_WENT_WRONG
         else:
             pip_install_archives_from(temp_path)
+
+            if satisfied_reqs:
+                print "These packages were already installed, so we didn't need to download or build"
+                print "them again. If you installed them with peep in the first place, you should be"
+                print "safe. If not, uninstall them, then re-attempt your install with peep."
+                for req in satisfied_reqs:
+                    print '   ', req.req
+
     return ITS_FINE_ITS_FINE
 
 
